## Brian Blaylock
## September 16, 2020

"""
=============
Retrieve Data
=============
Download and read data from the R-series Geostationary Operational
Environmental Satellite data.

Data is downloaded from Amazon Web Services and can be returned
as a file list or read as an xarray.Dataset. If the data is not
available in a local directory, it is loaded directly into memory.

https://registry.opendata.aws/noaa-goes/
"""

from os import pardir
from pathlib import Path
from datetime import datetime, timedelta
import multiprocessing

import s3fs
import xarray as xr
import pandas as pd
import numpy as np

# NOTE: These config dict values are retrieved from __init__ and read
# from the file ${HOME}/.config/goes2go/config.toml
from . import config

# Connect to AWS public buckets
fs = s3fs.S3FileSystem(anon=True)

# Define parameter options and aliases
# ------------------------------------
_satellite = {
    "noaa-goes16": [16, "16", "G16", "EAST", "GOES16"],
    "noaa-goes17": [17, "17", "G17", "WEST", "GOES17"],
}

_domain = {
    "C": ["CONUS"],
    "F": ["FULL", "FULLDISK", "FULL DISK"],
    "M": ["MESOSCALE", "M1", "M2"],
}

_product = {
    # Assume goes17 has same products as goes16
    i.split("/")[-1]: []
    for i in fs.ls(f"noaa-goes16")
}
_product.pop("index.html", None)
_product["GLM-L2-LCFA"] = ["GLM"]
_product["ABI-L2-MCMIPC"] = ["ABIC"]
_product["ABI-L2-MCMIPF"] = ["ABIF"]
_product["ABI-L2-MCMIPM"] = ["ABIM"]


def _check_param_inputs(**params):
    """
    Checks the input parameters for correct name or alias.

    Specifically, check the input for product, domain, and satellite are
    in the list of accepted values. If not, then look if it has an alias.
    """
    # Kinda messy, but gets the job done.
    params.setdefault("verbose", True)
    satellite = params["satellite"]
    domain = params["domain"]
    product = params["product"]
    verbose = params["verbose"]

    ## Determine the Satellite
    if satellite not in _satellite:
        satellite = str(satellite).upper()
        for key, aliases in _satellite.items():
            if satellite in aliases:
                satellite = key
    assert (
        satellite in _satellite
    ), f"satellite must be one of {list(_satellite.keys())} or an alias {list(_satellite.values())}"

    ## Determine the Domain (only needed for ABI product)
    if product.upper().startswith("ABI"):
        if product[-1] in _domain:
            # If the product has the domain, this takes priority
            domain = product[-1]
        elif isinstance(domain, str):
            domain = domain.upper()
            if domain in ["M1", "M2"]:
                product = product + "M"
            else:
                for key, aliases in _domain.items():
                    if domain in aliases:
                        domain = key
                product = product + domain
        assert (domain in _domain) or (
            domain in ["M1", "M2"]
        ), f"domain must be one of {list(_domain.keys())} or an alias {list(_domain.values())}"
    else:
        domain = None

    ## Determine the Product
    if product not in _product:
        for key, aliases in _product.items():
            if product.upper() in aliases:
                product = key
    assert (
        product in _product
    ), f"product must be one of {list(_product .keys())} or an alias {list(_product .values())}"

    if verbose:
        print(f" _______________________________")
        print(f" | Satellite: {satellite:<17s}|")
        print(f" |   Product: {product:<17s}|")
        if domain != None:
            print(f" |    Domain: {domain:<17s}|")

    return satellite, product, domain


def _goes_file_df(satellite, product, start, end, refresh=True):
    """
    Get list of requested GOES files as pandas.DataFrame.

    Parameters
    ----------
    satellite : str
    product : str
    start : datetime
    end : datetime
    refresh : bool
        Refresh the s3fs.S3FileSystem object when files are listed.
        Default True will refresh and not use a cached list.
    """
    params = locals()

    DATES = pd.date_range(f"{start:%Y-%m-%d %H:00}", f"{end:%Y-%m-%d %H:00}", freq="1H")

    # List all files for each date
    # ----------------------------
    files = []
    for DATE in DATES:
        files += fs.ls(f"{satellite}/{product}/{DATE:%Y/%j/%H/}", refresh=refresh)

    # Build a table of the files
    # --------------------------
    df = pd.DataFrame(files, columns=["file"])
    df[["start", "end", "creation"]] = (
        df["file"].str.rsplit("_", expand=True, n=3).loc[:, 1:]
    )

    # Filter files by requested time range
    # ------------------------------------
    # Convert filename datetime string to datetime object
    df["start"] = pd.to_datetime(df.start, format="s%Y%j%H%M%S%f")
    df["end"] = pd.to_datetime(df.end, format="e%Y%j%H%M%S%f")
    df["creation"] = pd.to_datetime(df.creation, format="c%Y%j%H%M%S%f.nc")

    # Filter by files within the requested time range
    df = df.loc[df.start >= start].loc[df.end <= end].reset_index(drop=True)

    for i in params:
        df.attrs[i] = params[i]

    return df


def _download_MP(src, save_dir, overwrite, i=1, n=1, verbose=True):
    """
    Download a single file -- a multiprocessing helper

    Parameters
    ----------
    src : str
        The source path of the file (path relative to S3 or ``save_dir``)
    save_dir : str or pathlib.Path
        Directory to save the file when it is downloaded.
    overwrite : bool
        True - overwrite files if the exist
        False - skip download if the file exists (default)
    i, n : int
        The iteration number and total number of files (primarily for
        the multiprocessor counter).
    """

    # File destination
    dst = Path(save_dir) / src

    if not dst.parent.is_dir():
        dst.parent.mkdir(parents=True, exist_ok=True)
        print(f"👨🏻‍🏭 Created directory: [{dst.parent}]", end="")

    if verbose:
        print(
            f"\r🏇🏻💨 Downloading ({i:,}/{n:,}) file from AWS to Local Disk. {src} > {dst}",
            end=" ",
        )

    if dst.is_file() and not overwrite:
        if verbose:
            print("---- 👮🏻‍♂️ File already exists. Do not overwrite.", end="")
    else:
        # Downloading file from AWS
        # NOTE: The destination is `str(dst)` and not just `dst` because
        #       Gaffney does not like a Path object for some reason.
        fs.get(src, str(dst))


def _download(df, **params):
    """
    Download each file in the pandas.DataFrame to ``save_dir``.

    Use multiprocessing to improve download time.

    Parameters
    ----------
    df : pandas.DataFrame
        A DataFrame created by ``_goes_file_df`` with a list of files
        to download from the GOES s3 bucket.
    save_dir : str or pathlib.Path
        Directory the files will be downloaded to.
        The default location (set as a variable in this script)
        is ``~/data/``.
    max_cpus : int or None
        If None, use all available CPUs. Else, specify the number of
        CPUs to use.
    overwrite : bool
        True - overwrite files if the exist
        False - skip download if the file exists (default)
    """
    params.setdefault("max_cpus", None)
    params.setdefault("verbose", True)
    save_dir = params["save_dir"]
    overwrite = params["overwrite"]
    max_cpus = params["max_cpus"]
    verbose = params["verbose"]

    n = len(df.file)
    if n == 0:
        print("🛸 No data to download.")
    elif n == 1:
        # If we only have one file, we don't need multiprocessing
        _download_MP(df.file[0], save_dir, overwrite, 1, 1, verbose)
    elif max_cpus == 1:
        [_download_MP(df.file[i], save_dir, overwrite, i, n, verbose) for i in range(n)]
    else:
        # Use Multiprocessing to download multiple files.
        if max_cpus is None:
            max_cpus = multiprocessing.cpu_count()
        cpus = np.minimum(multiprocessing.cpu_count(), max_cpus)
        cpus = np.minimum(cpus, n)

        inputs = [
            (src, save_dir, overwrite, i, n, verbose)
            for i, src in enumerate(df.file, start=1)
        ]

        with multiprocessing.Pool(cpus) as p:
            results = p.starmap(_download_MP, inputs)
            p.close()
            p.join()
    if verbose:
        print(
            f"\r{'':1000}\r📦 Finished downloading [{n}] files to [{save_dir/Path(df.file[0]).parents[3]}]."
        )


def _as_xarray_MP(src, save_dir, i=None, n=None, verbose=True):
    """Open a file as a xarray.Dataset -- a multiprocessing helper"""

    # File destination
    local_copy = Path(save_dir) / src

    if local_copy.is_file():
        if verbose:
            print(
                f"\r📖💽 Reading ({i:,}/{n:,}) file from LOCAL COPY [{local_copy}].",
                end=" ",
            )
        ds = xr.open_dataset(local_copy)
    else:
        if verbose:
            print(
                f"\r📖☁ Reading ({i:,}/{n:,}) file from AWS to MEMORY [{src}].", end=" "
            )
        with fs.open(src, "rb") as f:
            ds = xr.open_dataset(f).copy(deep=True)

    # Turn some attributes to coordinates so they will be preserved
    # when we concat multiple GOES DataSets together.
    attr2coord = [
        "dataset_name",
        "date_created",
        "time_coverage_start",
        "time_coverage_end",
    ]
    for i in attr2coord:
        if i in ds.attrs:
            ds.coords[i] = ds.attrs.pop(i)

    ds["filename"] = src

    return ds


def _as_xarray(df, **params):
    """
    Download files in the list to the desired path.

    Use multiprocessing to speed up the download process.

    Parameters
    ----------
    df : pandas.DataFrame
        A list of files in the GOES s3 bucket.
        This DataFrame must have a column of "files"
    params : dict
        Parameters from `goes_*` function.
    """
    params.setdefault("max_cpus", None)
    params.setdefault("verbose", True)
    save_dir = params["save_dir"]
    max_cpus = params["max_cpus"]
    verbose = params["verbose"]

    n = len(df.file)
    if n == 0:
        print("🛸 No data....🌌")
    elif n == 1:
        # If we only have one file, we don't need multiprocessing
        ds = _as_xarray_MP(df.iloc[0].file, save_dir, 1, 1, verbose)
    else:
        # Use Multiprocessing to read multiple files.
        if max_cpus is None:
            max_cpus = multiprocessing.cpu_count()
        cpus = np.minimum(multiprocessing.cpu_count(), max_cpus)
        cpus = np.minimum(cpus, n)

        inputs = [(src, save_dir, i, n) for i, src in enumerate(df.file, start=1)]

        with multiprocessing.Pool(cpus) as p:
            results = p.starmap(_as_xarray_MP, inputs)
            p.close()
            p.join()

        # Need some work to concat the datasets
        if df.attrs["product"].startswith("ABI"):
            print("concatenate Datasets", end="")
            ds = xr.concat(results, dim="t")
        else:
            ds = results

    if verbose:
        print(f"\r{'':1000}\r📚 Finished reading [{n}] files into xarray.Dataset.")
    ds.attrs["path"] = df.file.to_list()
    return ds


###############################################################################
###############################################################################


def goes_timerange(
    start=None,
    end=None,
    recent=None,
    *,
    satellite=config["timerange"].get("satellite"),
    product=config["timerange"].get("product"),
    domain=config["timerange"].get("domain"),
    return_as=config["timerange"].get("return_as"),
    download=config["timerange"].get("download"),
    overwrite=config["timerange"].get("overwrite"),
    save_dir=config["timerange"].get("save_dir"),
    max_cpus=config["timerange"].get("max_cpus"),
    s3_refresh=config["timerange"].get("s3_refresh"),
    verbose=config["timerange"].get("verbose", True),
):
    """
    Get GOES data for a time range.

    Parameters
    ----------
    start, end : datetime
        Required if recent is None.
    recent : timedelta or pandas-parsable timedelta str
        Required if start and end are None. If timedelta(hours=1), will
        get the most recent files for the past hour.
    satellite : {'goes16', 'goes17'}
        Specify which GOES satellite.
        The following alias may also be used:

        - ``'goes16'``: 16, 'G16', or 'EAST'
        - ``'goes17'``: 17, 'G17', or 'WEST'

    product : {'ABI', 'GLM', other GOES product}
        Specify the product name.

        - 'ABI' is an alias for ABI-L2-MCMIP Multichannel Cloud and Moisture Imagery
        - 'GLM' is an alias for GLM-L2-LCFA Geostationary Lightning Mapper

        Others may include ``'ABI-L1b-Rad'``, ``'ABI-L2-DMW'``, etc.
        For more available products, look at this `README
        <https://docs.opendata.aws/noaa-goes16/cics-readme.html>`_
    domain : {'C', 'F', 'M'}
        ABI scan region indicator. Only required for ABI products if the
        given product does not end with C, F, or M.

        - C: Contiguous United States (alias 'CONUS')
        - F: Full Disk (alias 'FULL')
        - M: Mesoscale (alias 'MESOSCALE')

    return_as : {'xarray', 'filelist'}
        Return the data as an xarray.Dataset or as a list of files
    download : bool
        - True: Download the data to disk to the location set by :guilabel:`save_dir`
        - False: Just load the data into memory.
    save_dir : pathlib.Path or str
        Path to save the data.
    overwrite : bool
        - True: Download the file even if it exists.
        - False Do not download the file if it already exists
    max_cpus : int
    s3_refresh : bool
        Refresh the s3fs.S3FileSystem object when files are listed.

    """
    # If `start`, or `end` is a string, parse with Pandas
    if isinstance(start, str):
        start = pd.to_datetime(start)
    if isinstance(end, str):
        end = pd.to_datetime(end)
    # If `recent` is a string (like recent='1H'), parse with Pandas
    if isinstance(recent, str):
        recent = pd.to_timedelta(recent)

    params = locals()
    satellite, product, domain = _check_param_inputs(**params)
    params["satellite"] = satellite
    params["product"] = product
    params["domain"] = domain

    check1 = start is not None and end is not None
    check2 = recent is not None
    assert check1 or check2, "🤔 `start` and `end` *or* `recent` is required"

    if check1:
        assert hasattr(start, "second") and hasattr(
            end, "second"
        ), "`start` and `end` must be a datetime object"
    elif check2:
        assert hasattr(recent, "seconds"), "`recent` must be a timedelta object"

    # Parameter Setup
    # ---------------
    # Create a range of directories to check. The GOES S3 bucket is
    # organized by hour of day.
    if recent is not None:
        start = datetime.utcnow() - recent
        end = datetime.utcnow()

    df = _goes_file_df(satellite, product, start, end, refresh=s3_refresh)

    if download:
        _download(df, **params)

    if return_as == "filelist":
        df.attrs["filePath"] = save_dir
        return df
    elif return_as == "xarray":
        return _as_xarray(df, **params)


def goes_latest(
    *,
    satellite=config["latest"].get("satellite"),
    product=config["latest"].get("product"),
    domain=config["latest"].get("domain"),
    return_as=config["latest"].get("return_as"),
    download=config["latest"].get("download"),
    overwrite=config["latest"].get("overwrite"),
    save_dir=config["latest"].get("save_dir"),
    s3_refresh=config["latest"].get("s3_refresh"),
    verbose=config["latest"].get("verbose", True),
):
    """
    Get the latest available GOES data.

    Parameters
    ----------
    satellite : {'goes16', 'goes17'}
        Specify which GOES satellite.
        The following alias may also be used:

        - ``'goes16'``: 16, 'G16', or 'EAST'
        - ``'goes17'``: 17, 'G17', or 'WEST'

    product : {'ABI', 'GLM', other GOES product}
        Specify the product name.

        - 'ABI' is an alias for ABI-L2-MCMIP Multichannel Cloud and Moisture Imagery
        - 'GLM' is an alias for GLM-L2-LCFA Geostationary Lightning Mapper

        Others may include ``'ABI-L1b-Rad'``, ``'ABI-L2-DMW'``, etc.
        For more available products, look at this `README
        <https://docs.opendata.aws/noaa-goes16/cics-readme.html>`_
    domain : {'C', 'F', 'M'}
        ABI scan region indicator. Only required for ABI products if the
        given product does not end with C, F, or M.

        - C: Contiguous United States (alias 'CONUS')
        - F: Full Disk (alias 'FULL')
        - M: Mesoscale (alias 'MESOSCALE')

    return_as : {'xarray', 'filelist'}
        Return the data as an xarray.Dataset or as a list of files
    download : bool
        - True: Download the data to disk to the location set by :guilabel:`save_dir`
        - False: Just load the data into memory.
    save_dir : pathlib.Path or str
        Path to save the data.
    overwrite : bool
        - True: Download the file even if it exists.
        - False Do not download the file if it already exists
    s3_refresh : bool
        Refresh the s3fs.S3FileSystem object when files are listed.
    """
    params = locals()
    satellite, product, domain = _check_param_inputs(**params)
    params["satellite"] = satellite
    params["product"] = product
    params["domain"] = domain

    # Parameter Setup
    # ---------------
    # Create a range of directories to check. The GOES S3 bucket is
    # organized by hour of day. Look in the current hour and last hour.
    start = datetime.utcnow() - timedelta(hours=1)
    end = datetime.utcnow()

    df = _goes_file_df(satellite, product, start, end, refresh=s3_refresh)

    # Filter for specific mesoscale domain
    if domain is not None and domain.upper() in ["M1", "M2"]:
        df = df[df["file"].str.contains(f"{domain.upper()}-M")]

    # Get the most recent file (latest start date)
    df = df.loc[df.start == df.start.max()].reset_index(drop=True)

    if download:
        _download(df, **params)

    if return_as == "filelist":
        df.attrs["filePath"] = save_dir
        return df
    elif return_as == "xarray":
        return _as_xarray(df, **params)


def goes_nearesttime(
    attime,
    within=pd.to_timedelta(config["nearesttime"].get("within", "1H")),
    *,
    satellite=config["nearesttime"].get("satellite"),
    product=config["nearesttime"].get("product"),
    domain=config["nearesttime"].get("domain"),
    return_as=config["nearesttime"].get("return_as"),
    download=config["nearesttime"].get("download"),
    overwrite=config["nearesttime"].get("overwrite"),
    save_dir=config["nearesttime"].get("save_dir"),
    s3_refresh=config["nearesttime"].get("s3_refresh"),
    verbose=config["nearesttime"].get("verbose", True),
):
    """
    Get the latest available GOES data.

    Parameters
    ----------
    attime : datetime
        Time to find the nearest observation for.
        May also use a pandas-interpretable datetime string.
    within : timedelta or pandas-parsable timedelta str
        Timerange tht the nearest observation must be.
    satellite : {'goes16', 'goes17'}
        Specify which GOES satellite.
        The following alias may also be used:

        - ``'goes16'``: 16, 'G16', or 'EAST'
        - ``'goes17'``: 17, 'G17', or 'WEST'

    product : {'ABI', 'GLM', other GOES product}
        Specify the product name.

        - 'ABI' is an alias for ABI-L2-MCMIP Multichannel Cloud and Moisture Imagery
        - 'GLM' is an alias for GLM-L2-LCFA Geostationary Lightning Mapper

        Others may include ``'ABI-L1b-Rad'``, ``'ABI-L2-DMW'``, etc.
        For more available products, look at this `README
        <https://docs.opendata.aws/noaa-goes16/cics-readme.html>`_
    domain : {'C', 'F', 'M'}
        ABI scan region indicator. Only required for ABI products if the
        given product does not end with C, F, or M.

        - C: Contiguous United States (alias 'CONUS')
        - F: Full Disk (alias 'FULL')
        - M: Mesoscale (alias 'MESOSCALE')

    return_as : {'xarray', 'filelist'}
        Return the data as an xarray.Dataset or as a list of files
    download : bool
        - True: Download the data to disk to the location set by :guilabel:`save_dir`
        - False: Just load the data into memory.
    save_dir : pathlib.Path or str
        Path to save the data.
    overwrite : bool
        - True: Download the file even if it exists.
        - False: Do not download the file if it already exists
    s3_refresh : bool
        Refresh the s3fs.S3FileSystem object when files are listed.
    """
    if isinstance(attime, str):
        attime = pd.to_datetime(attime)
    if isinstance(within, str):
        within = pd.to_timedelta(within)

    params = locals()
    satellite, product, _ = _check_param_inputs(**params)
    params["satellite"] = satellite
    params["product"] = product

    # Parameter Setup
    # ---------------
    # Create a range of directories to check. The GOES S3 bucket is
    # organized by hour of day.
    start = attime - within
    end = attime + within

    df = _goes_file_df(satellite, product, start, end, refresh=s3_refresh)

    # return df, start, end, attime

    # Filter for specific mesoscale domain
    if domain.upper() in ["M1", "M2"]:
        df = df[df["file"].str.contains(f"{domain.upper()}-M")]

    # Get row that matches the nearest time
    df = df.sort_values("start")
    df = df.set_index(df.start)
    nearest_time_index = df.index.get_loc(attime, method="nearest")
    df = df.iloc[[nearest_time_index]]
    df = df.reset_index(drop=True)

    n = len(df.file)
    if n == 0:
        print("🛸 No data....🌌")
        return None

    if download:
        _download(df, **params)

    if return_as == "filelist":
        df.attrs["filePath"] = save_dir
        return df
    elif return_as == "xarray":
        return _as_xarray(df, **params)



## Brian Blaylock
## August 8, 2019

"""
===========
RGB Recipes
===========

.. image:: /_static/RGB_sample.png

These functions take GOES-East or GOES-West multichannel data on a 
fixed grid (files named ``ABI-L2-MCMIPC``) and generates a 3D 
Red-Green-Blue (RGB) array for various GOES RGB products.

RGB recipes are based on the `GOES Quick Guides
<http://rammb.cira.colostate.edu/training/visit/quick_guides/>`_ 
and include the following:

    - NaturalColor
    - TrueColor
    - FireTemperature
    - AirMass
    - DayCloudPhase
    - DayConvection
    - DayCloudConvection
    - DayLandCloud
    - DayLandCloudFire
    - WaterVapor
    - DifferentialWaterVapor
    - DaySnowFog
    - NighttimeMicrophysics
    - Dust
    - SulfurDioxide
    - Ash
    - SplitWindowDifference
    - NightFogDifference
    - RocketPlume              ✨New - July 9, 2021

The returned RGB can easily be viewed with ``plt.imshow(RGB)``. 

For imshow to show an RGB image, the values must range between 0 and 1. 
Values are normalized between the range specified in the Quick Guides. 
This normalization is synonymous to `contrast or histogram stretching 
<https://micro.magnet.fsu.edu/primer/java/digitalimaging/processing/histogramstretching/index.html>`_
(`more info here
<https://staff.fnwi.uva.nl/r.vandenboomgaard/IPCV20162017/LectureNotes/IP/PointOperators/ImageStretching.html>`_)
and follows the formula:

    .. code-block:: python 

        NormalizedValue = (OriginalValue-LowerLimit)/(UpperLimit-LowerLimit)

`Gamma correction <https://en.wikipedia.org/wiki/Gamma_correction>`_
darkens or lightens an image (`more info 
<https://www.cambridgeincolour.com/tutorials/gamma-correction.htm>`_) 
and follows the decoding formula:

    .. code-block:: python 
        
        R_corrected = R**(1/gamma)

The input for all these functions are denoted by ``C`` for "channels" which
represents the GOES ABI multichannel file opened with xarray. For example:

    .. code-block:: python 
        
        FILE = 'OR_ABI-L2-MCMIPC-M6_G17_s20192201631196_e20192201633575_c20192201634109.nc'
        C = xarray.open_dataset(FILE)

All RGB products are demonstarted in the `make_RGB_Demo 
<https://github.com/blaylockbk/goes2go/tree/master/notebooks>`_ notebook.

Note: I don't have a `GeoColor <https://journals.ametsoc.org/view/journals/atot/37/3/JTECH-D-19-0134.1.xml>`_
RGB, because it is much more involved than simply stacking RGB channels. If anyone does do
something similar to a GeoColor image, let me know!

ABI Band Reference
------------------

https://www.weather.gov/media/crp/GOES_16_Guides_FINALBIS.pdf
http://cimss.ssec.wisc.edu/goes/GOESR_QuickGuides.html
https://www.goes-r.gov/mission/ABI-bands-quick-info.html

=============== ================== ============================================== ======================================
ABI Band Number Central Wavelength  Name                                          Type
=============== ================== ============================================== ======================================
      1             0.47 μm        "Blue" Band                                    Visible
      2             0.64 μm        "Red" Band                                     Visible
      3             0.86 μm        "Veggie" Band                                  Near-IR
      4             1.37 μm        "Cirrus" Band                                  Near-IR
      5             1.6  μm        "Snow/Ice" Band                                Near-IR
      6             2.2  μm        "Cloud Particle Size" Band                     Near-IR
      7             3.9  μm        "Shortwave Window" Band                        IR (with reflected daytime component)
      8             6.2  μm        "Upper-Level Tropospheric Water Vapor" Band    IR
      9             6.9  μm        "Mid-Level Tropospheric Water Vapor" Band      IR
      10            7.3  μm        "Lower-level Water Vapor" Band                 IR
      11            8.4  μm        "Cloud-Top Phase" Band                         IR
      12            9.6  μm        "Ozone Band"                                   IR
      13            10.3 μm        "Clean" IR Longwave Window Band                IR
      14            11.2 μm        IR Longwave Window Band                        IR
      15            12.3 μm        "Dirty" Longwave Window Band                   IR
      16            13.3 μm        "CO2" Longwave infrared                        IR
=============== ================== ============================================== ======================================


"""

import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import xarray as xr

from goes2go.tools import field_of_view


def get_imshow_kwargs(ds):
    """
    Help determine the ``plt.imshow`` arguments.

    Parameters
    ----------
    ds : xarray.Dataset

    Returns
    -------
    kwargs for the ``plt.imshow`` with the correct image extent limits.

    Examples
    --------

    .. code:: python

        r = TrueColor(G)
        ax = common_features(r.crs)
        ax.imshow(r.TrueColor, *\*\get_imshow_kwargs(r))

    """
    return dict(
        extent=[ds.x2.data.min(), ds.x2.data.max(), ds.y2.data.min(), ds.y2.data.max()],
        transform=ds.crs,
        origin="upper",
        interpolation="none",
    )


def rgb_as_dataset(G, RGB, description, latlon=False):
    """
    Assemble a dataset with the RGB array with other data from the file.

    Parameters
    ----------
    G : xarray.Dataset
        GOES ABI data from multispectral channel
    RGB : array
        A 3D array of R, G, and B values at each pixel
    description : str
        A description of what the RGB data represents.
    latlon : bool
        Derive the latitude and longitude of each pixel.

    """
    # Assemble a new xarray.Dataset for the RGB data
    ds = xr.Dataset({description.replace(" ", ""): (["y", "x", "rgb"], RGB)})
    ds.attrs["description"] = description

    # Convert x, y points to latitude/longitude
    _, crs = field_of_view(G)
    sat_h = G.goes_imager_projection.perspective_point_height
    x2 = G.x * sat_h
    y2 = G.y * sat_h
    ds.coords["x2"] = x2
    ds.coords["y2"] = y2

    ds["x2"].attrs["long_name"] = "x sweep in crs units (m); x * sat_height"
    ds["y2"].attrs["long_name"] = "y sweep in crs units (m); y * sat_height"

    ds.attrs["crs"] = crs

    if latlon:
        X, Y = np.meshgrid(x2, y2)
        a = ccrs.PlateCarree().transform_points(crs, X, Y)
        lons, lats, _ = a[:, :, 0], a[:, :, 1], a[:, :, 2]
        ds.coords["longitude"] = (("y", "x"), lons)
        ds.coords["latitude"] = (("y", "x"), lats)

    # Copy some coordinates and attributes of interest from the original data
    for i in ["x", "y", "t", "geospatial_lat_lon_extent"]:
        ds.coords[i] = G[i]
    for i in [
        "orbital_slot",
        "platform_ID",
        "scene_id",
        "spatial_resolution",
        "instrument_type",
        "title",
    ]:
        ds.attrs[i] = G.attrs[i]

    ## Provide some helpers to plot with imshow
    ds.attrs["imshow_kwargs"] = get_imshow_kwargs(ds)

    ## Provide some helpers to plot with imshow and pcolormesh
    ## Not super useful, because pcolormesh doesn't allow nans in x, y dimension
    # pcolormesh_kwargs = dict(
    #    color = RGB.reshape(np.shape(RGB)[0] * np.shape(RGB)[1], np.shape(RGB)[2])
    #    shading='nearest'
    #    )
    # ds.attrs['pcolormesh_kwargs'] = pcolormesh_kwargs

    return ds


def load_RGB_channels(C, channels):
    """
    Return the R, G, and B arrays for the three channels requested. This
    function will convert the data any units in Kelvin to Celsius.

    Parameters
    ----------
    C : xarray.Dataset
        The GOES multi-channel file opened with xarray.
    channels : tuple of size 3
        A tuple of the channel number for each (R, G, B).
        For example ``channel=(2, 3, 1)`` is for the true color RGB

    Returns
    -------
    A list with three items that are used for R, G, and B.
    >>> R, G, B = load_RGB_channels(C, (2,3,1))

    """
    # Units of each channel requested
    units = [C["CMI_C%02d" % c].units for c in channels]
    RGB = []
    for u, c in zip(units, channels):
        if u == "K":
            # Convert form Kelvin to Celsius
            RGB.append(C["CMI_C%02d" % c].data - 273.15)
        else:
            RGB.append(C["CMI_C%02d" % c].data)
    return RGB


def gamma_correction(a, gamma, verbose=False):
    """
    Darken or lighten an image with `gamma correction
    <https://en.wikipedia.org/wiki/Gamma_correction>`_.

    Parameters
    ----------
    a : array-like
        An array of values, typically the RGB array of values in
        an image.
    gamma : float
        Gamma value to decode the image by.
        Values > 1 will lighten an image.
        Values < 1 will darken an image.
    """
    if verbose:
        if gamma > 1:
            print("Gamma Correction: 🌔 Lighten image")
        elif gamma < 1:
            print("Gamma Correction: 🌒 Darken image")
        else:
            print("Gamma Correction: 🌓 Gamma=1. No correction made.")
            return a

    # Gamma decoding formula
    return np.power(a, 1 / gamma)


def normalize(value, lower_limit, upper_limit, clip=True):
    """
    Normalize values between 0 and 1.

    Normalize between a lower and upper limit. In other words, it
    converts your number to a value in the range between 0 and 1.
    Follows `normalization formula
    <https://stats.stackexchange.com/a/70807/220885>`_

    This is the same concept as `contrast or histogram stretching
    <https://staff.fnwi.uva.nl/r.vandenboomgaard/IPCV20162017/LectureNotes/IP/PointOperators/ImageStretching.html>`_


    .. code:: python

        NormalizedValue = (OriginalValue-LowerLimit)/(UpperLimit-LowerLimit)

    Parameters
    ----------
    value :
        The original value. A single value, vector, or array.
    upper_limit :
        The upper limit.
    lower_limit :
        The lower limit.
    clip : bool
        - True: Clips values between 0 and 1 for RGB.
        - False: Retain the numbers that extends outside 0-1 range.
    Output:
        Values normalized between the upper and lower limit.
    """
    norm = (value - lower_limit) / (upper_limit - lower_limit)
    if clip:
        norm = np.clip(norm, 0, 1)
    return norm


# ======================================================================
# ======================================================================


def TrueColor(C, gamma=2.2, pseudoGreen=True, night_IR=True, **kwargs):
    """
    True Color RGB:
    (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf>`__ for reference)

    This is similar to the NaturalColor RGB, but uses a different gamma
    correction and does not apply contrast stretching. I think these
    images look a little "washed out" when compared to the NaturalColor
    RGB. So, I would recommend using the NaturalColor RGB.

    For more details on combing RGB and making the psedo green channel, refer to
    `Bah et al. 2018 <https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2018EA000379>`_.

    .. image:: /_static/TrueColor.png

    .. image:: /_static/gamma_demo_TrueColor.png

    .. image:: /_static/Color-IR_demo.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    gamma : float
        Darken or lighten an image with `gamma correction
        <https://en.wikipedia.org/wiki/Gamma_correction>`_.
        Values > 1 will lighten an image.
        Values < 1 will darken an image.
    pseudoGreen : bool
        True: returns the calculated "True" green color
        False: returns the "veggie" channel
    night_IR : bool
        If True, use Clean IR (channel 13) as maximum RGB value overlay
        so that cold clouds show up at night. (Be aware that some
        daytime clouds might appear brighter).
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (2, 3, 1))

    # Apply range limits for each channel. RGB values must be between 0 and 1
    R = np.clip(R, 0, 1)
    G = np.clip(G, 0, 1)
    B = np.clip(B, 0, 1)

    # Apply a gamma correction to each R, G, B channel
    R = gamma_correction(R, gamma)
    G = gamma_correction(G, gamma)
    B = gamma_correction(B, gamma)

    if pseudoGreen:
        # Calculate the "True" Green
        G = 0.45 * R + 0.1 * G + 0.45 * B
        G = np.clip(G, 0, 1)

    if night_IR:
        # Load the Clean IR channel
        IR = C["CMI_C13"]
        # Normalize between a range and clip
        IR = normalize(IR, 90, 313, clip=True)
        # Invert colors so cold clouds are white
        IR = 1 - IR
        # Lessen the brightness of the coldest clouds so they don't
        # appear so bright when we overlay it on the true color image
        IR = IR / 1.4
        # RGB with IR as greyscale
        RGB = np.dstack([np.maximum(R, IR), np.maximum(G, IR), np.maximum(B, IR)])
    else:
        RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "True Color", **kwargs)


def NaturalColor(C, gamma=0.8, pseudoGreen=True, night_IR=False, **kwargs):
    """
    Natural Color RGB based on CIMSS method. Thanks Rick Kohrs!
    (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf>`__ for reference)

    Check out Rick Kohrs `merged GOES images <https://www.ssec.wisc.edu/~rickk/local-noon.html>`_.

    This NaturalColor RGB is *very* similar to the TrueColor RGB but
    uses slightly different contrast stretches and ranges.

    For more details on combing RGB and making the psedo green channel, refer to
    `Bah et al. 2018 <https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2018EA000379>`_.

    .. image:: /_static/NaturalColor.png

    .. image:: /_static/gamma_demo_NaturalColor-PsuedoGreen.png

    .. image:: /_static/gamma_demo_NaturalColor-VeggieGreen.png

    .. image:: /_static/Color-IR_demo.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened ith xarray.
    gamma : float
        Darken or lighten an image with `gamma correction
        <https://en.wikipedia.org/wiki/Gamma_correction>`_.
        Values > 1 will lighten an image.
        Values < 1 will darken an image.
    night_IR : bool
        If True, use Clean IR (channel 13) as maximum RGB value overlay
        so that cold clouds show up at night. (Be aware that some
        daytime clouds might appear brighter).
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel
    """

    def breakpoint_stretch(C, breakpoint):
        """
        Contrast stretching by break point (number provided by Rick Kohrs)
        """
        lower = normalize(C, 0, 10)  # Low end
        upper = normalize(C, 10, 255)  # High end

        # Combine the two datasets
        # This works because if upper=1 and lower==.7, then
        # that means the upper value was out of range and the
        # value for the lower pass was used instead.
        combined = np.minimum(lower, upper)

        return combined

    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (2, 3, 1))

    # Apply range limits for each channel. RGB values must be between 0 and 1
    R = np.clip(R, 0, 1)
    G = np.clip(G, 0, 1)
    B = np.clip(B, 0, 1)

    if pseudoGreen:
        # Derive pseudo Green channel
        G = 0.45 * R + 0.1 * G + 0.45 * B
        G = np.clip(G, 0, 1)

    # Convert Albedo to Brightness, ranging from 0-255 K
    # (numbers based on email from Rick Kohrs)
    R = np.sqrt(R * 100) * 25.5
    G = np.sqrt(G * 100) * 25.5
    B = np.sqrt(B * 100) * 25.5

    # Apply contrast stretching based on breakpoints
    # (numbers based on email form Rick Kohrs)
    R = breakpoint_stretch(R, 33)
    G = breakpoint_stretch(G, 40)
    B = breakpoint_stretch(B, 50)

    if night_IR:
        # Load the Clean IR channel
        IR = C["CMI_C13"]
        # Normalize between a range and clip
        IR = normalize(IR, 90, 313, clip=True)
        # Invert colors so cold clouds are white
        IR = 1 - IR
        # Lessen the brightness of the coldest clouds so they don't
        # appear so bright when we overlay it on the true color image
        IR = IR / 1.4
        # Overlay IR channel, as greyscale image (use IR in R, G, and B)
        RGB = np.dstack([np.maximum(R, IR), np.maximum(G, IR), np.maximum(B, IR)])
    else:
        RGB = np.dstack([R, G, B])

    # Apply a gamma correction to the image
    RGB = gamma_correction(RGB, gamma)

    return rgb_as_dataset(C, RGB, "Natural Color", **kwargs)


def FireTemperature(C, **kwargs):
    """
    Fire Temperature RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Fire_Temperature_RGB.pdf>`__ for reference)

    .. image:: /_static/FireTemperature.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (7, 6, 5))

    # Normalize each channel by the appropriate range of values (clipping happens in function)
    R = normalize(R, 273, 333)
    G = normalize(G, 0, 1)
    B = normalize(B, 0, 0.75)

    # Apply the gamma correction to Red channel.
    #   corrected_value = value^(1/gamma)
    gamma = 0.4
    R = gamma_correction(R, gamma)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Fire Temperature", **kwargs)


def AirMass(C, **kwargs):
    """
    Air Mass RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_AirMassRGB_final.pdf>`__ for reference)

    .. image:: /_static/AirMass.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C08"].data - C["CMI_C10"].data
    G = C["CMI_C12"].data - C["CMI_C13"].data
    B = C["CMI_C08"].data - 273.15  # remember to convert to Celsius

    # Normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
    R = normalize(R, -26.2, 0.6)
    G = normalize(G, -42.2, 6.7)
    B = normalize(B, -64.65, -29.25)

    # Invert B
    B = 1 - B

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Air Mass", **kwargs)


def DayCloudPhase(C, **kwargs):
    """
    Day Cloud Phase Distinction RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Day_Cloud_Phase_Distinction.pdf>`__ for reference)

    .. image:: /_static/DayCloudPhase.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (13, 2, 5))

    # Normalize each channel by the appropriate range of values. (Clipping happens inside function)
    R = normalize(R, -53.5, 7.5)
    G = normalize(G, 0, 0.78)
    B = normalize(B, 0.01, 0.59)

    # Invert R
    R = 1 - R

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Cloud Phase", **kwargs)


def DayConvection(C, **kwargs):
    """
    Day Convection RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayConvectionRGB_final.pdf>`__ for reference)

    .. image:: /_static/DayConvection.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    # NOTE: Each R, G, B is a channel difference.
    R = C["CMI_C08"].data - C["CMI_C10"].data
    G = C["CMI_C07"].data - C["CMI_C13"].data
    B = C["CMI_C05"].data - C["CMI_C02"].data

    # Normalize each channel by the appropriate range of values.
    R = normalize(R, -35, 5)
    G = normalize(G, -5, 60)
    B = normalize(B, -0.75, 0.25)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Convection", **kwargs)


def DayCloudConvection(C, **kwargs):
    """
    Day Cloud Convection RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DayCloudConvectionRGB_final.pdf>`__ for reference)

    .. image:: /_static/DayCloudConvection.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (2, 2, 13))

    # Normalize each channel by the appropriate range of values.
    R = normalize(R, 0, 1)
    G = normalize(G, 0, 1)
    B = normalize(B, -70.15, 49.85)

    # Invert B
    B = 1 - B

    # Apply the gamma correction to Red channel.
    #   corrected_value = value^(1/gamma)
    gamma = 1.7
    R = gamma_correction(R, gamma)
    G = gamma_correction(G, gamma)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Cloud Convection", **kwargs)


def DayLandCloud(C, **kwargs):
    """
    Day Land Cloud Fire RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_daylandcloudRGB_final.pdf>`__ for reference)

    .. image:: /_static/DayLandCloud.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (5, 3, 2))

    # Normalize each channel by the appropriate range of values  e.g. R = (R-minimum)/(maximum-minimum)
    R = normalize(R, 0, 0.975)
    G = normalize(G, 0, 1.086)
    B = normalize(B, 0, 1)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Land Cloud", **kwargs)


def DayLandCloudFire(C, **kwargs):
    """
    Day Land Cloud Fire RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayLandCloudFireRGB_final.pdf>`__ for reference)

    .. image:: /_static/DayLandCloudFire.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R, G, B = load_RGB_channels(C, (6, 3, 2))

    # Normalize each channel by the appropriate range of values  e.g. R = (R-minimum)/(maximum-minimum)
    R = normalize(R, 0, 1)
    G = normalize(G, 0, 1)
    B = normalize(B, 0, 1)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Land Cloud Fire", **kwargs)


def WaterVapor(C, **kwargs):
    """
    Simple Water Vapor RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Simple_Water_Vapor_RGB.pdf>`__ for reference)

    .. image:: /_static/WaterVapor.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables.
    R, G, B = load_RGB_channels(C, (13, 8, 10))

    # Normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
    R = normalize(R, -70.86, 5.81)
    G = normalize(G, -58.49, -30.48)
    B = normalize(B, -28.03, -12.12)

    # Invert the colors
    R = 1 - R
    G = 1 - G
    B = 1 - B

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Water Vapor", **kwargs)


def DifferentialWaterVapor(C, **kwargs):
    """
    Differential Water Vapor RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DifferentialWaterVaporRGB_final.pdf>`__ for reference)

    .. image:: /_static/DifferentialWaterVapor.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables.
    R = C["CMI_C10"].data - C["CMI_C08"].data
    G = C["CMI_C10"].data - 273.15
    B = C["CMI_C08"].data - 273.15

    # Normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
    R = normalize(R, -3, 30)
    G = normalize(G, -60, 5)
    B = normalize(B, -64.65, -29.25)

    # Gamma correction
    R = gamma_correction(R, 0.2587)
    G = gamma_correction(G, 0.4)
    B = gamma_correction(B, 0.4)

    # Invert the colors
    R = 1 - R
    G = 1 - G
    B = 1 - B

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Differenctial Water Vapor", **kwargs)


def DaySnowFog(C, **kwargs):
    """
    Day Snow-Fog RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DaySnowFog.pdf>`__ for reference)

    .. image:: /_static/DaySnowFog.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C03"].data
    G = C["CMI_C05"].data
    B = C["CMI_C07"].data - C["CMI_C13"].data

    # Normalize values
    R = normalize(R, 0, 1)
    G = normalize(G, 0, 0.7)
    B = normalize(B, 0, 30)

    # Apply a gamma correction to the image
    gamma = 1.7
    R = gamma_correction(R, gamma)
    G = gamma_correction(G, gamma)
    B = gamma_correction(B, gamma)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Day Snow Fog", **kwargs)


def NighttimeMicrophysics(C, **kwargs):
    """
    Nighttime Microphysics RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_NtMicroRGB_final.pdf>`__ for reference)

    .. image:: /_static/NighttimeMicrophysics.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C15"].data - C["CMI_C13"].data
    G = C["CMI_C13"].data - C["CMI_C07"].data
    B = C["CMI_C13"].data - 273.15

    # Normalize values
    R = normalize(R, -6.7, 2.6)
    G = normalize(G, -3.1, 5.2)
    B = normalize(B, -29.6, 19.5)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Nighttime Microphysics", **kwargs)


def Dust(C, **kwargs):
    """
    SulfurDioxide RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Dust_RGB_Quick_Guide.pdf>`__ for reference)

    .. image:: /_static/Dust.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C15"].data - C["CMI_C13"].data
    G = C["CMI_C14"].data - C["CMI_C11"].data
    B = C["CMI_C13"].data - 273.15

    # Normalize values
    R = normalize(R, -6.7, 2.6)
    G = normalize(G, -0.5, 20)
    B = normalize(B, -11.95, 15.55)

    # Apply a gamma correction to the image
    gamma = 2.5
    G = gamma_correction(G, gamma)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Dust", **kwargs)


def SulfurDioxide(C, **kwargs):
    """
    SulfurDioxide RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Quick_Guide_SO2_RGB.pdf>`__ for reference)

    .. image:: /_static/SulfurDioxide.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C09"].data - C["CMI_C10"].data
    G = C["CMI_C13"].data - C["CMI_C11"].data
    B = C["CMI_C07"].data - 273.15

    # Normalize values
    R = normalize(R, -4, 2)
    G = normalize(G, -4, 5)
    B = normalize(B, -30.1, 29.8)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Sulfur Dioxide", **kwargs)


def Ash(C, **kwargs):
    """
    Ash RGB:
    (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/GOES_Ash_RGB.pdf>`__ for reference)

    .. image:: /_static/Ash.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C15"].data - C["CMI_C13"].data
    G = C["CMI_C14"].data - C["CMI_C11"].data
    B = C["CMI_C13"].data - 273.15

    # Normalize values
    R = normalize(R, -6.7, 2.6)
    G = normalize(G, -6, 6.3)
    B = normalize(B, -29.55, 29.25)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "Ash", **kwargs)


def SplitWindowDifference(C, **kwargs):
    """
    Split Window Difference RGB (greyscale):
    (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_SplitWindowDifference.pdf>`__ for reference)

    .. image:: /_static/SplitWindowDifference.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    data = C["CMI_C15"].data - C["CMI_C13"].data

    # Normalize values
    data = normalize(data, -10, 10)

    # The final RGB array :)
    RGB = np.dstack([data, data, data])

    return rgb_as_dataset(C, RGB, "Split Window Difference", **kwargs)


def NightFogDifference(C, **kwargs):
    """
    Night Fog Difference RGB (greyscale):
    (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_NightFogBTD.pdf>`__ for reference)

    .. image:: /_static/NightFogDifference.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    data = C["CMI_C13"].data - C["CMI_C07"].data

    # Normalize values
    data = normalize(data, -90, 15)

    # Invert data
    data = 1 - data

    # The final RGB array :)
    RGB = np.dstack([data, data, data])

    return rgb_as_dataset(C, RGB, "Night Fog Difference", **kwargs)


def RocketPlume(C, night=False, **kwargs):
    """
    Rocket Plume RGB

    For identifying rocket launches.

    See `this blog <https://cimss.ssec.wisc.edu/satellite-blog/archives/41335>`__ and
    the `Quick Guide <https://cimss.ssec.wisc.edu/satellite-blog/images/2021/06/QuickGuide_Template_GOESRBanner_Rocket_Plume.pdf>`__
    for reference

    .. image:: /_static/RocketPlume.png

    Parameters
    ----------
    C : xarray.Dataset
        A GOES ABI multichannel file opened with xarray.
    night : bool
        If the area is in night, turn this on to use a different channel
        than the daytime application.
    \*\*kwargs :
        Keyword arguments for ``rgb_as_dataset`` function.
        - latlon : derive latitude and longitude of each pixel

    """
    # Load the three channels into appropriate R, G, and B variables
    R = C["CMI_C07"].data
    G = C["CMI_C08"].data
    if not night:
        B = C["CMI_C02"].data
    else:
        B = C["CMI_C05"].data

    # Normalize values
    R = normalize(R, 273, 338)
    G = normalize(G, 233, 253)
    B = normalize(B, 0, 0.80)

    # The final RGB array :)
    RGB = np.dstack([R, G, B])

    return rgb_as_dataset(C, RGB, "RocketPlume", **kwargs)


#######################
# 🚧 Construction Zone
#######################
def NormalizedBurnRatio(C, **kwargs):
    """
    Normalized Burn Ratio

    **THIS FUNCTION IS NOT FULLY DEVELOPED. Need more info.**

    NBR= (0.86 µm – 2.2 µm)/(0.86 um + 2.2 um)


    https://ntrs.nasa.gov/citations/20190030825

    Parameters
    ----------

    """
    # Load the three channels into appropriate R, G, and B variables
    C3 = C["CMI_C03"].data
    C6 = C["CMI_C06"].data
    data = (C3 - C6) / (C3 + C6)

    # Invert data
    # data = 1-data

    # The final RGB array :)
    RGB = np.dstack([data, data, data])

    return rgb_as_dataset(C, RGB, "Normalized Burn Ratio", **kwargs)


if __name__ == "__main__":

    # Create images of each for Docs
    print("nothing here for now")



## Brian Blaylock
## July 8, 2021

import warnings
import toml
from pathlib import Path
import os

########################################################################
# Load custom xarray accessors
# TODO: Move some of the tools.py to these accessors.
try:
   import goes2go.accessors
except:
   warnings.warn("goes2go xarray accessors could not be imported.")

########################################################################
# Append Path object with my custom expand method so user can use
# environment variables in the config file (e.g., ${HOME}).
def _expand(self):
    """
    Fully expand and resolve the Path with the given environment variables.

    Example
    -------
    >>> Path('$HOME').expand()
    >>> PosixPath('/p/home/blaylock')
    """
    return Path(os.path.expandvars(self)).expanduser().resolve()


Path.expand = _expand

########################################################################
# goes2go configuration file
# Configuration file is save in `~/config/goes2go/config.toml`
_config_path = Path("~/.config/goes2go/config.toml").expand()
_save_dir = str(Path('~/data').expand())

# NOTE: The `\\` is an escape character in TOML.
# For Windows paths "C:\\user\\"" needs to be "C:\\\\user\\\\""
_save_dir = str(Path('~/data').expand())
_save_dir = _save_dir.replace('\\', '\\\\')

########################################################################
# Default TOML Configuration
default_toml = f"""
["default"]
save_dir = "{_save_dir}"
satellite = "noaa-goes16"
product = "ABI-L2-MCMIP"
domain = "C"
download = true
return_as = "filelist"
overwrite = false
max_cpus = 1
s3_refresh = true
verbose = true

["timerange"]
s3_refresh = false

["latest"]
return_as = "xarray"

["nearesttime"]
within = "1H"
return_as = "xarray"
"""

########################################################################
# If a config file isn't found, make one
if not _config_path.exists():
    _config_path.parent.mkdir(parents=True, exist_ok=True)
    with open(_config_path, "w") as f:
        toml_string = toml.dump(toml.loads(default_toml), f)
    print(f"⚙ Created config file [{_config_path}] with default values.")

########################################################################
# Read the config file
config = toml.load(_config_path)

config["default"]["save_dir"] = Path(config["default"]["save_dir"]).expand()

# Merge default settings with overwrite settings for each download method
for i in ["timerange", "latest", "nearesttime"]:
    config[i] = {**config["default"], **config[i]}



## Brian Blaylock
## September 22, 2020

"""
===========
Other Tools
===========
Other tools for handeling NOAA GOES data files.
"""

import numpy as np
import cartopy.crs as ccrs
import warnings

try:
    import metpy  # Need accessors to get projection info.
except:
    # Not sure why sphinx can't import metpy??
    warnings.warn('Metpy not imported.')
from shapely.geometry import Point, Polygon


def field_of_view(G, resolution=60, reduce_abi_fov=0.06):
    """
    Create a field-of-view polygon for the GOES data.

    Based on information from the `GOES-R Series Data Book
    <https://www.goes-r.gov/downloads/resources/documents/GOES-RSeriesDataBook.pdf>`_.

    GLM lense field of view is 16 degree, or +/- 8 degrees (see page 225)
    ABI full-disk field of view if 17.4 degrees (see page 48)

    To plot the field of view on the cartopy axes, do the following:

    .. code:: python

        FOV, geo = field_of_view(G)
        ax = plt.subplot(projection=geo)
        ax.add_geometries([FOV], crs=geo)

    Parameters
    ----------
    G : xarray.Dataset
        The GOES NetCDF file opened with xarray. A file is required
        because we get info from the file to define the projection.
    resolution : int
        Resolution of polygon shapes
    reduce_abi_fov : float or int
        Since the globe isn't a perfect ellipse, reduce the field of
        view just slightly to get all the points to be on the projection
        plane. If this number is less than the default, the polygon
        will not be calculated correctly because edge points will lie
        off the projection globe.
    """
    warnings.warn('DEPRECIATION. Use the FOV accessor instead `G.FOV.full_disk` or `G.FOV.domain`')
    if G.title.startswith("ABI"):
        globe_kwargs = dict(
            semimajor_axis=G.goes_imager_projection.semi_major_axis,
            semiminor_axis=G.goes_imager_projection.semi_minor_axis,
            inverse_flattening=G.goes_imager_projection.inverse_flattening,
        )
        sat_height = G.goes_imager_projection.perspective_point_height
        nadir_lon = G.geospatial_lat_lon_extent.geospatial_lon_nadir
        nadir_lat = G.geospatial_lat_lon_extent.geospatial_lat_nadir
        # Field of view in degrees
        FOV = 17.4
        FOV -= reduce_abi_fov  # little less to account for imprecise ellipsoid
    elif G.title.startswith("GLM"):
        globe_kwargs = dict(
            semimajor_axis=G.goes_lat_lon_projection.semi_major_axis,
            semiminor_axis=G.goes_lat_lon_projection.semi_minor_axis,
            inverse_flattening=G.goes_lat_lon_projection.inverse_flattening,
        )
        sat_height = G.nominal_satellite_height.item() * 1000
        nadir_lon = G.lon_field_of_view.item()
        nadir_lat = G.lat_field_of_view.item()
        FOV = 8 * 2
        FOV += 0.15  # Little offset to better match boundary from Rudlosky et al. 2018

    # Create a cartopy coordinate reference system for the data

    # These numbers are from the `goes_imager_projection` variable
    globe = ccrs.Globe(ellipse=None, **globe_kwargs)

    crs = ccrs.Geostationary(
        central_longitude=nadir_lon,
        satellite_height=sat_height,
        globe=globe,
        sweep_axis="x",
    )

    # Create polygon of the field of view. This polygon is in
    # the geostationary crs projection units, and is in meters.
    # The central point is at 0,0 (not the nadir position), because
    # we are working in the geostationary projection coordinates
    # and the center point is 0,0 meters.
    FOV_radius = np.radians(FOV / 2) * sat_height
    FOV_poly = Point(0, 0).buffer(FOV_radius, resolution=resolution)

    ## GLM is a bit funny. I haven't found this in the documentation
    ## anywhere, yet, but the GLM field-of-view is not exactly
    ## the full circle, there is a square area cut out of it.
    ## The square FOV is ~ 15 degrees
    if G.title.startswith("GLM"):
        FOV_square = 15 / 2
        # FOV_square += .1 # offset to match Rudlosky et al. 2018
        FOV_radius = np.radians(FOV_square) * sat_height
        # Create a square with many points clockwise, starting in bottom left corner
        side1x, side1y = (
            np.ones(resolution) * -FOV_radius,
            np.linspace(-FOV_radius, FOV_radius, resolution),
        )
        side2x, side2y = (
            np.linspace(-FOV_radius, FOV_radius, resolution),
            np.ones(resolution) * FOV_radius,
        )
        side3x, side3y = (
            np.ones(resolution) * FOV_radius,
            np.linspace(FOV_radius, -FOV_radius, resolution),
        )
        side4x, side4y = (
            np.linspace(FOV_radius, -FOV_radius, resolution),
            np.ones(resolution) * -FOV_radius,
        )
        x = np.hstack([side1x, side2x, side3x, side4x])
        y = np.hstack([side1y, side2y, side3y, side4y])
        square_FOV = Polygon(zip(x, y))
        FOV_poly = FOV_poly.intersection(square_FOV)

        return FOV_poly, crs

    if G.title.startswith("ABI"):
        # We have the global field of view
        # now we need the domain field of view
        dom_border = np.array(
            [(i, G.y.data[0]) for i in G.x.data]
            + [(G.x.data[-1], i) for i in G.y.data]
            + [(i, G.y.data[-1]) for i in G.x.data[::-1]]
            + [(G.x.data[0], i) for i in G.y.data[::-1]]
        )
        FOV_dom = Polygon(dom_border * sat_height)
        FOV_dom = FOV_dom.intersection(FOV_poly)
        return FOV_poly, FOV_dom, crs


def abi_crs(G, reference_variable="CMI_C01"):
    """
    Get coordinate reference system for the Advanced Baseline Imager (ABI).

    Parameters
    ----------
    G : xarray.Dataset
        An xarray.Dataset to derive the coordinate reference system.
    reference_variable : str
        A variable in the xarray.Dataset to use to parse projection from.

    Returns
    -------
    Three objects are returned
        1. cartopy coordinate reference system
        2. data projection coordinates in x direction
        3. data projection coordinates in y direction

    """
    # We'll use the `CMI_C01` variable as a 'hook' to get the CF metadata.
    dat = G.metpy.parse_cf(reference_variable)

    crs = dat.metpy.cartopy_crs

    # We also need the x (north/south) and y (east/west) axis sweep of the ABI data
    x, y = (dat.x, dat.y)

    return crs, x, y


def glm_crs(G, reference_variable="flash_lat"):
    """Not too useful, because it's just lat/lon coordinates"""
    dat = G.metpy.parse_cf("flash_lat")
    crs = dat.metpy.cartopy_crs
    return crs



## Brian Blaylock
## July 9, 2021

"""
===========
RGB Recipes
===========

RGB Recipes for the GOES Advanced Baseline Imager.

More about xarray accessors:
http://xarray.pydata.org/en/stable/internals/extending-xarray.html?highlight=extending#
"""

import warnings
import xarray as xr
import numpy as np
import cartopy.crs as ccrs

from shapely.geometry import Point, Polygon

########################
# Image Processing Tools
def _gamma_correction(a, gamma, verbose=False):
    """
    Darken or lighten an image with `gamma correction
    <https://en.wikipedia.org/wiki/_gamma_correction>`_.

    Parameters
    ----------
    a : array-like
        An array of values, typically the RGB array of values in
        an image.
    gamma : float
        Gamma value to decode the image by.
        Values > 1 will lighten an image.
        Values < 1 will darken an image.
    """
    if verbose:
        if gamma > 1:
            print("Gamma Correction: 🌔 Lighten image")
        elif gamma < 1:
            print("Gamma Correction: 🌒 Darken image")
        else:
            print("Gamma Correction: 🌓 Gamma=1. No correction made.")
            return a

    # Gamma decoding formula
    return np.power(a, 1 / gamma)


def _normalize(value, lower_limit, upper_limit, clip=True):
    """
    Normalize values between an upper and lower limit between 0 and 1.

    Normalize between a lower and upper limit. In other words, it
    converts your number to a value in the range between 0 and 1.
    Follows `normalization formula
    <https://stats.stackexchange.com/a/70807/220885>`_

    This is the same concept as `contrast or histogram stretching
    <https://staff.fnwi.uva.nl/r.vandenboomgaard/IPCV20162017/LectureNotes/IP/PointOperators/ImageStretching.html>`_


    .. code:: python

        _normalizedValue = (OriginalValue-LowerLimit)/(UpperLimit-LowerLimit)

    Parameters
    ----------
    value :
        The original value. A single value, vector, or array.
    upper_limit :
        The upper limit.
    lower_limit :
        The lower limit.
    clip : bool
        - True: Clips values between 0 and 1 for RGB.
        - False: Retain the numbers that extends outside 0-1 range.
    """
    norm = (value - lower_limit) / (upper_limit - lower_limit)
    if clip:
        norm = np.clip(norm, 0, 1)
    return norm


@xr.register_dataset_accessor("FOV")
class fieldOfViewAccessor:
    """
    Create a field-of-view polygon for the GOES data.

    Based on information from the `GOES-R Series Data Book
    <https://www.goes-r.gov/downloads/resources/documents/GOES-RSeriesDataBook.pdf>`_.

    GLM lense field of view is 16 degree, or +/- 8 degrees (see page 225)
    ABI full-disk field of view if 17.4 degrees (see page 48)
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    @property
    def crs(self):
        """Cartopy coordinate reference system for the Satellite."""
        ds = self._obj
        if ds.title.startswith("ABI"):
            globe_kwargs = dict(
                semimajor_axis=ds.goes_imager_projection.semi_major_axis,
                semiminor_axis=ds.goes_imager_projection.semi_minor_axis,
                inverse_flattening=ds.goes_imager_projection.inverse_flattening,
            )
            sat_height = ds.goes_imager_projection.perspective_point_height
            nadir_lon = ds.geospatial_lat_lon_extent.geospatial_lon_nadir
            nadir_lat = ds.geospatial_lat_lon_extent.geospatial_lat_nadir
        elif ds.title.startswith("GLM"):
            globe_kwargs = dict(
                semimajor_axis=ds.goes_lat_lon_projection.semi_major_axis,
                semiminor_axis=ds.goes_lat_lon_projection.semi_minor_axis,
                inverse_flattening=ds.goes_lat_lon_projection.inverse_flattening,
            )
            sat_height = ds.nominal_satellite_height.item() * 1000
            nadir_lon = ds.lon_field_of_view.item()
            nadir_lat = ds.lat_field_of_view.item()
        # Create a cartopy coordinate reference system (crs)
        globe = ccrs.Globe(ellipse=None, **globe_kwargs)

        crs = ccrs.Geostationary(
            central_longitude=nadir_lon,
            satellite_height=sat_height,
            globe=globe,
            sweep_axis="x",
        )

        return crs

    @property
    def full_disk(self):
        """
        Full-disk field of view for the ABI or GLM instruments.

        .. image:: /_static/ABI_field-of-view.png

        .. image:: /_static/GLM_field-of-view.png

        Returns
        -------
        shapely.Polygon
        """
        ds = self._obj

        # Create polygon of the field of view. This polygon is in
        # the geostationary crs projection units, and is in meters.
        # The central point is at 0,0 (not the nadir position), because
        # we are working in the geostationary projection coordinates
        # and the center point is 0,0 meters.
        if ds.title.startswith("ABI"):
            # Field of view (FOV) in degrees. Reduce just a little to
            # get all polygon points in the projection plane so cartopy
            # can plot it correctly.
            # TODO: Is there a more "correct" way to handle this?
            sat_height = ds.goes_imager_projection.perspective_point_height
            FOV_degrees = 17.4
            FOV_degrees -= 0.06
            FOV_radius = np.radians(FOV_degrees / 2) * sat_height
            FOV_polygon = Point(0, 0).buffer(FOV_radius, resolution=160)
            return FOV_polygon
        elif ds.title.startswith("GLM"):
            # Field of view (FOV) of GLM is different than ABI.
            # Do a little offset to better match boundary from
            # Rudlosky et al. 2018
            sat_height = ds.nominal_satellite_height.item() * 1000
            FOV_degrees = 8 * 2
            FOV_degrees += 0.15
            FOV_radius = np.radians(FOV_degrees / 2) * sat_height
            FOV_polygon = Point(0, 0).buffer(FOV_radius, resolution=160)

            # I haven't found this explained in the documentation yet,
            # but the GLM field-of-view is not exactly the full circle,
            # there is a square area cut out of it.
            # The square FOV width and height is about 15 degrees.
            cutout_FOV_degrees = 15 / 2
            cutout_FOV_length = np.radians(cutout_FOV_degrees) * sat_height
            # Create a square with many points clockwise, starting in bottom left corner
            side_points = np.linspace(-cutout_FOV_length, cutout_FOV_length, 300)
            cutout_points = np.array(
                [(-cutout_FOV_length, i) for i in side_points]
                + [(i, cutout_FOV_length) for i in side_points]
                + [(cutout_FOV_length, i) for i in side_points][::-1]
                + [(i, -cutout_FOV_length) for i in side_points][::-1]
            )
            cutout = Polygon(cutout_points)
            FOV_polygon = FOV_polygon.intersection(cutout)
            return FOV_polygon

    @property
    def domain(self):
        """
        Field of view for the ABI domain (CONUS or MesoScale).

        .. image:: /_static/ABI_field-of-view_16dom.png

        .. image:: /_static/ABI_field-of-view_16M1M2.png

        .. image:: /_static/ABI_field-of-view_17dom.png

        Returns
        -------
        shapely.Polygon
        """
        ds = self._obj
        assert ds.title.startswith(
            "ABI"
        ), "Domain polygon only available for ABI CONUS and Mesoscale files."
        sat_height = ds.goes_imager_projection.perspective_point_height
        # Trim out domain FOV from the full disk (this is necessary for GOES-16).
        dom_border = np.array(
            [(i, ds.y.data[0]) for i in ds.x.data]
            + [(ds.x.data[-1], i) for i in ds.y.data]
            + [(i, ds.y.data[-1]) for i in ds.x.data[::-1]]
            + [(ds.x.data[0], i) for i in ds.y.data[::-1]]
        )
        FOV_domain = Polygon(dom_border * sat_height)
        FOV_domain = FOV_domain.intersection(self.full_disk)
        return FOV_domain


@xr.register_dataset_accessor("rgb")
class rgbAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        assert (
            self._obj.title == "ABI L2 Cloud and Moisture Imagery"
        ), "Dataset must be an ABI L2 Cloud and Moisture Imagery file."
        self._crs = None
        self._x = None
        self._y = None
        self._sat_h = self._obj.goes_imager_projection.perspective_point_height
        self._imshow_kwargs = None

    @property
    def crs(self):
        """Cartopy coordinate reference system"""
        if self._crs is None:
            # Why am I doing this? To Cache the values.
            self._crs = self._obj.FOV.crs
        return self._crs

    @property
    def x(self):
        """x sweep in crs units (m); x * sat_height"""
        if self._x is None:
            self._x = self._obj.x * self._sat_h
        return self._x

    @property
    def y(self):
        """y sweep in crs units (m); x * sat_height"""
        if self._y is None:
            self._y = self._obj.y * self._sat_h
        return self._y

    @property
    def imshow_kwargs(self):
        """Key word arguments for plt.imshow for generating images.

        Projection axis must be the coordinate reference system.
        """
        if self._imshow_kwargs is None:
            self._imshow_kwargs = dict(
                extent=[
                    self.x.data.min(),
                    self.x.data.max(),
                    self.y.data.min(),
                    self.y.data.max(),
                ],
                transform=self._crs,
                origin="upper",
                interpolation="none",
            )
        return self._imshow_kwargs

    def get_latlon(self):
        """Get lat/lon of all points"""
        X, Y = np.meshgrid(self.x, self.y)
        a = ccrs.PlateCarree().transform_points(self._crs, X, Y)
        lons, lats, _ = a[:, :, 0], a[:, :, 1], a[:, :, 2]

        self._obj.coords["longitude"] = (("y", "x"), lons)
        self._obj.coords["latitude"] = (("y", "x"), lats)
        return self._obj["latitude"], self._obj["longitude"]

    ####################################################################
    # Helpers
    def _load_RGB_channels(self, channels):
        """
        Return the R, G, and B arrays for the three channels requested. This
        function will convert the data any units in Kelvin to Celsius.

        Parameters
        ----------
        channels : tuple of size 3
            A tuple of the channel number for each (R, G, B).
            For example ``channel=(2, 3, 1)`` is for the true color RGB

        Returns
        -------
        A list with three items that are used for R, G, and B.
        >>> R, G, B = _load_RGB_channels((2,3,1))

        """
        ds = self._obj

        # Units of each channel requested
        units = [ds["CMI_C%02d" % c].units for c in channels]
        RGB = []
        for u, c in zip(units, channels):
            if u == "K":
                # Convert form Kelvin to Celsius                                ## <-- Do I REALLY want to hard-code this in?
                RGB.append(ds["CMI_C%02d" % c].data - 273.15)
            else:
                RGB.append(ds["CMI_C%02d" % c].data)
        return RGB

    ####################################################################
    # RGB Recipes
    def TrueColor(self, gamma=2.2, pseudoGreen=True, night_IR=True):
        """
        True Color RGB:
        (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf>`__ for reference)

        This is similar to the NaturalColor RGB, but uses a different gamma
        correction and does not apply contrast stretching. I think these
        images look a little "washed out" when compared to the NaturalColor
        RGB. So, I would recommend using the NaturalColor RGB.

        For more details on combing RGB and making the psedo green channel, refer to
        `Bah et al. 2018 <https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2018EA000379>`_.

        .. image:: /_static/TrueColor.png

        .. image:: /_static/gamma_demo_TrueColor.png

        .. image:: /_static/Color-IR_demo.png

        Parameters
        ----------
        gamma : float
            Darken or lighten an image with `gamma correction
            <https://en.wikipedia.org/wiki/_gamma_correction>`_.
            Values > 1 will lighten an image.
            Values < 1 will darken an image.
        pseudoGreen : bool
            True: returns the calculated "True" green color
            False: returns the "veggie" channel
        night_IR : bool
            If True, use Clean IR (channel 13) as maximum RGB value overlay
            so that cold clouds show up at night. (Be aware that some
            daytime clouds might appear brighter).


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((2, 3, 1))

        # Apply range limits for each channel. RGB values must be between 0 and 1
        R = np.clip(R, 0, 1)
        G = np.clip(G, 0, 1)
        B = np.clip(B, 0, 1)

        # Apply a gamma correction to each R, G, B channel
        R = _gamma_correction(R, gamma)
        G = _gamma_correction(G, gamma)
        B = _gamma_correction(B, gamma)

        if pseudoGreen:
            # Calculate the "True" Green
            G = 0.45 * R + 0.1 * G + 0.45 * B
            G = np.clip(G, 0, 1)

        if night_IR:
            # Load the Clean IR channel
            IR = ds["CMI_C13"]
            # _normalize between a range and clip
            IR = _normalize(IR, 90, 313, clip=True)
            # Invert colors so cold clouds are white
            IR = 1 - IR
            # Lessen the brightness of the coldest clouds so they don't
            # appear so bright when we overlay it on the true color image
            IR = IR / 1.4
            # RGB with IR as greyscale
            RGB = np.dstack([np.maximum(R, IR), np.maximum(G, IR), np.maximum(B, IR)])
        else:
            RGB = np.dstack([R, G, B])

        ds["TrueColor"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["TrueColor"].attrs[
            "Quick Guide"
        ] = "http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf"
        ds["TrueColor"].attrs["long_name"] = "True Color"

        return ds["TrueColor"]

    def NaturalColor(self, gamma=0.8, pseudoGreen=True, night_IR=False):
        """
        Natural Color RGB based on CIMSS method. Thanks Rick Kohrs!
        (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf>`__ for reference)

        Check out Rick Kohrs `merged GOES images <https://www.ssec.wisc.edu/~rickk/local-noon.html>`_.

        This NaturalColor RGB is *very* similar to the TrueColor RGB but
        uses slightly different contrast stretches and ranges.

        For more details on combing RGB and making the psedo green channel, refer to
        `Bah et al. 2018 <https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2018EA000379>`_.

        .. image:: /_static/NaturalColor.png

        .. image:: /_static/gamma_demo_NaturalColor-PsuedoGreen.png

        .. image:: /_static/gamma_demo_NaturalColor-VeggieGreen.png

        .. image:: /_static/Color-IR_demo.png

        Parameters
        ----------
        gamma : float
            Darken or lighten an image with `gamma correction
            <https://en.wikipedia.org/wiki/_gamma_correction>`_.
            Values > 1 will lighten an image.
            Values < 1 will darken an image.
        night_IR : bool
            If True, use Clean IR (channel 13) as maximum RGB value overlay
            so that cold clouds show up at night. (Be aware that some
            daytime clouds might appear brighter).

        """
        ds = self._obj

        def breakpoint_stretch(C, breakpoint):
            """
            Contrast stretching by break point (number provided by Rick Kohrs)
            """
            lower = _normalize(C, 0, 10)  # Low end
            upper = _normalize(C, 10, 255)  # High end

            # Combine the two datasets
            # This works because if upper=1 and lower==.7, then
            # that means the upper value was out of range and the
            # value for the lower pass was used instead.
            combined = np.minimum(lower, upper)

            return combined

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((2, 3, 1))

        # Apply range limits for each channel. RGB values must be between 0 and 1
        R = np.clip(R, 0, 1)
        G = np.clip(G, 0, 1)
        B = np.clip(B, 0, 1)

        if pseudoGreen:
            # Derive pseudo Green channel
            G = 0.45 * R + 0.1 * G + 0.45 * B
            G = np.clip(G, 0, 1)

        # Convert Albedo to Brightness, ranging from 0-255 K
        # (numbers based on email from Rick Kohrs)
        R = np.sqrt(R * 100) * 25.5
        G = np.sqrt(G * 100) * 25.5
        B = np.sqrt(B * 100) * 25.5

        # Apply contrast stretching based on breakpoints
        # (numbers based on email form Rick Kohrs)
        R = breakpoint_stretch(R, 33)
        G = breakpoint_stretch(G, 40)
        B = breakpoint_stretch(B, 50)

        if night_IR:
            # Load the Clean IR channel
            IR = ds["CMI_C13"]
            # _normalize between a range and clip
            IR = _normalize(IR, 90, 313, clip=True)
            # Invert colors so cold clouds are white
            IR = 1 - IR
            # Lessen the brightness of the coldest clouds so they don't
            # appear so bright when we overlay it on the true color image
            IR = IR / 1.4
            # Overlay IR channel, as greyscale image (use IR in R, G, and B)
            RGB = np.dstack([np.maximum(R, IR), np.maximum(G, IR), np.maximum(B, IR)])
        else:
            RGB = np.dstack([R, G, B])

        # Apply a gamma correction to the image
        RGB = _gamma_correction(RGB, gamma)

        ds["NaturalColor"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["NaturalColor"].attrs[
            "Quick Guide"
        ] = "http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_CIMSSRGB_v2.pdf"
        ds["NaturalColor"].attrs["long_name"] = "Natural Color"

        return ds["NaturalColor"]

    def FireTemperature(self):
        """
        Fire Temperature RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Fire_Temperature_RGB.pdf>`__ for reference)

        .. image:: /_static/FireTemperature.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((7, 6, 5))

        # _normalize each channel by the appropriate range of values (clipping happens in function)
        R = _normalize(R, 273, 333)
        G = _normalize(G, 0, 1)
        B = _normalize(B, 0, 0.75)

        # Apply the gamma correction to Red channel.
        #   corrected_value = value^(1/gamma)
        gamma = 0.4
        R = _gamma_correction(R, gamma)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["FireTemperature"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["FireTemperature"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/Fire_Temperature_RGB.pdf"
        ds["FireTemperature"].attrs["long_name"] = "Fire Temperature"

        return ds["FireTemperature"]

    def AirMass(self):
        """
        Air Mass RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_AirMassRGB_final.pdf>`__ for reference)

        .. image:: /_static/AirMass.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C08"].data - ds["CMI_C10"].data
        G = ds["CMI_C12"].data - ds["CMI_C13"].data
        B = ds["CMI_C08"].data - 273.15  # remember to convert to Celsius

        # _normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
        R = _normalize(R, -26.2, 0.6)
        G = _normalize(G, -42.2, 6.7)
        B = _normalize(B, -64.65, -29.25)

        # Invert B
        B = 1 - B

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["AirMass"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["AirMass"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_AirMassRGB_final.pdf"
        ds["AirMass"].attrs["long_name"] = "Air Mass"

        return ds["AirMass"]

    def DayCloudPhase(self):
        """
        Day Cloud Phase Distinction RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Day_Cloud_Phase_Distinction.pdf>`__ for reference)

        .. image:: /_static/DayCloudPhase.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((13, 2, 5))

        # _normalize each channel by the appropriate range of values. (Clipping happens inside function)
        R = _normalize(R, -53.5, 7.5)
        G = _normalize(G, 0, 0.78)
        B = _normalize(B, 0.01, 0.59)

        # Invert R
        R = 1 - R

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DayCloudPhase"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DayCloudPhase"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/Day_Cloud_Phase_Distinction.pdf"
        ds["DayCloudPhase"].attrs["long_name"] = "Day Cloud Phase"

        return ds["DayCloudPhase"]

    def DayConvection(self):
        """
        Day Convection RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayConvectionRGB_final.pdf>`__ for reference)

        .. image:: /_static/DayConvection.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        # NOTE: Each R, G, B is a channel difference.
        R = ds["CMI_C08"].data - ds["CMI_C10"].data
        G = ds["CMI_C07"].data - ds["CMI_C13"].data
        B = ds["CMI_C05"].data - ds["CMI_C02"].data

        # _normalize each channel by the appropriate range of values.
        R = _normalize(R, -35, 5)
        G = _normalize(G, -5, 60)
        B = _normalize(B, -0.75, 0.25)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DayConvection"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DayConvection"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayConvectionRGB_final.pdf"
        ds["DayConvection"].attrs["long_name"] = "Day Convection"

        return ds["DayConvection"]

    def DayCloudConvection(self):
        """
        Day Cloud Convection RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DayCloudConvectionRGB_final.pdf>`__ for reference)

        .. image:: /_static/DayCloudConvection.png

        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((2, 2, 13))

        # _normalize each channel by the appropriate range of values.
        R = _normalize(R, 0, 1)
        G = _normalize(G, 0, 1)
        B = _normalize(B, -70.15, 49.85)

        # Invert B
        B = 1 - B

        # Apply the gamma correction to Red channel.
        #   corrected_value = value^(1/gamma)
        gamma = 1.7
        R = _gamma_correction(R, gamma)
        G = _gamma_correction(G, gamma)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DayCloudConvection"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DayCloudConvection"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DayCloudConvectionRGB_final.pdf"
        ds["DayCloudConvection"].attrs["long_name"] = "Day Cloud Convection"

        return ds["DayCloudConvection"]

    def DayLandCloud(self):
        """
        Day Land Cloud Fire RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_daylandcloudRGB_final.pdf>`__ for reference)

        .. image:: /_static/DayLandCloud.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((5, 3, 2))

        # _normalize each channel by the appropriate range of values  e.g. R = (R-minimum)/(maximum-minimum)
        R = _normalize(R, 0, 0.975)
        G = _normalize(G, 0, 1.086)
        B = _normalize(B, 0, 1)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DayLandCloud"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DayLandCloud"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_daylandcloudRGB_final.pdf"
        ds["DayLandCloud"].attrs["long_name"] = "Day Land Cloud"

        return ds["DayLandCloud"]

    def DayLandCloudFire(self):
        """
        Day Land Cloud Fire RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayLandCloudFireRGB_final.pdf>`__ for reference)

        .. image:: /_static/DayLandCloudFire.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R, G, B = self._load_RGB_channels((6, 3, 2))

        # _normalize each channel by the appropriate range of values  e.g. R = (R-minimum)/(maximum-minimum)
        R = _normalize(R, 0, 1)
        G = _normalize(G, 0, 1)
        B = _normalize(B, 0, 1)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DayLandCloudFire"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DayLandCloudFire"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DayLandCloudFireRGB_final.pdf"
        ds["DayLandCloudFire"].attrs["long_name"] = "Day Land Cloud Fire"

        return ds["DayLandCloud"]

    def WaterVapor(self):
        """
        Simple Water Vapor RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Simple_Water_Vapor_RGB.pdf>`__ for reference)

        .. image:: /_static/WaterVapor.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables.
        R, G, B = self._load_RGB_channels((13, 8, 10))

        # _normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
        R = _normalize(R, -70.86, 5.81)
        G = _normalize(G, -58.49, -30.48)
        B = _normalize(B, -28.03, -12.12)

        # Invert the colors
        R = 1 - R
        G = 1 - G
        B = 1 - B

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["WaterVapor"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["WaterVapor"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/Simple_Water_Vapor_RGB.pdf"
        ds["WaterVapor"].attrs["long_name"] = "Water Vapor"

        return ds["WaterVapor"]

    def DifferentialWaterVapor(self):
        """
        Differential Water Vapor RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DifferentialWaterVaporRGB_final.pdf>`__ for reference)

        .. image:: /_static/DifferentialWaterVapor.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables.
        R = ds["CMI_C10"].data - ds["CMI_C08"].data
        G = ds["CMI_C10"].data - 273.15
        B = ds["CMI_C08"].data - 273.15

        # _normalize each channel by the appropriate range of values. e.g. R = (R-minimum)/(maximum-minimum)
        R = _normalize(R, -3, 30)
        G = _normalize(G, -60, 5)
        B = _normalize(B, -64.65, -29.25)

        # Gamma correction
        R = _gamma_correction(R, 0.2587)
        G = _gamma_correction(G, 0.4)
        B = _gamma_correction(B, 0.4)

        # Invert the colors
        R = 1 - R
        G = 1 - G
        B = 1 - B

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DifferentialWaterVapor"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DifferentialWaterVapor"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_DifferentialWaterVaporRGB_final.pdf"
        ds["DifferentialWaterVapor"].attrs["long_name"] = "Differential Water Vapor"

        return ds["DifferentialWaterVapor"]

    def DaySnowFog(self):
        """
        Day Snow-Fog RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DaySnowFog.pdf>`__ for reference)

        .. image:: /_static/DaySnowFog.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C03"].data
        G = ds["CMI_C05"].data
        B = ds["CMI_C07"].data - ds["CMI_C13"].data

        # _normalize values
        R = _normalize(R, 0, 1)
        G = _normalize(G, 0, 0.7)
        B = _normalize(B, 0, 30)

        # Apply a gamma correction to the image
        gamma = 1.7
        R = _gamma_correction(R, gamma)
        G = _gamma_correction(G, gamma)
        B = _gamma_correction(B, gamma)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["DaySnowFog"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["DaySnowFog"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_DaySnowFog.pdf"
        ds["DaySnowFog"].attrs["long_name"] = "Day Snow Fog"

        return ds["DaySnowFog"]

    def NighttimeMicrophysics(self):
        """
        Nighttime Microphysics RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_NtMicroRGB_final.pdf>`__ for reference)

        .. image:: /_static/NighttimeMicrophysics.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C15"].data - ds["CMI_C13"].data
        G = ds["CMI_C13"].data - ds["CMI_C07"].data
        B = ds["CMI_C13"].data - 273.15

        # _normalize values
        R = _normalize(R, -6.7, 2.6)
        G = _normalize(G, -3.1, 5.2)
        B = _normalize(B, -29.6, 19.5)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["NighttimeMicrophysics"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["NighttimeMicrophysics"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/QuickGuide_GOESR_NtMicroRGB_final.pdf"
        ds["NighttimeMicrophysics"].attrs["long_name"] = "Nighttime Microphysics"

        return ds["NighttimeMicrophysics"]

    def Dust(self):
        """
        SulfurDioxide RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Dust_RGB_Quick_Guide.pdf>`__ for reference)

        .. image:: /_static/Dust.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C15"].data - ds["CMI_C13"].data
        G = ds["CMI_C14"].data - ds["CMI_C11"].data
        B = ds["CMI_C13"].data - 273.15

        # _normalize values
        R = _normalize(R, -6.7, 2.6)
        G = _normalize(G, -0.5, 20)
        B = _normalize(B, -11.95, 15.55)

        # Apply a gamma correction to the image
        gamma = 2.5
        G = _gamma_correction(G, gamma)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["Dust"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["Dust"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/Dust_RGB_Quick_Guide.pdf"
        ds["Dust"].attrs["long_name"] = "Dust"

        return ds["Dust"]

    def SulfurDioxide(self):
        """
        SulfurDioxide RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/Quick_Guide_SO2_RGB.pdf>`__ for reference)

        .. image:: /_static/SulfurDioxide.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C09"].data - ds["CMI_C10"].data
        G = ds["CMI_C13"].data - ds["CMI_C11"].data
        B = ds["CMI_C07"].data - 273.15

        # _normalize values
        R = _normalize(R, -4, 2)
        G = _normalize(G, -4, 5)
        B = _normalize(B, -30.1, 29.8)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["SulfurDioxide"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["SulfurDioxide"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/Quick_Guide_SO2_RGB.pdf"
        ds["SulfurDioxide"].attrs["long_name"] = "Sulfur Dioxide"

        return ds["SulfurDioxide"]

    def Ash(self):
        """
        Ash RGB:
        (See `Quick Guide <http://rammb.cira.colostate.edu/training/visit/quick_guides/GOES_Ash_RGB.pdf>`__ for reference)

        .. image:: /_static/Ash.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C15"].data - ds["CMI_C13"].data
        G = ds["CMI_C14"].data - ds["CMI_C11"].data
        B = ds["CMI_C13"].data - 273.15

        # _normalize values
        R = _normalize(R, -6.7, 2.6)
        G = _normalize(G, -6, 6.3)
        B = _normalize(B, -29.55, 29.25)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["Ash"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["Ash"].attrs[
            "Quick Guide"
        ] = "http://rammb.cira.colostate.edu/training/visit/quick_guides/GOES_Ash_RGB.pdf"
        ds["Ash"].attrs["long_name"] = "Ash"

        return ds["Ash"]

    def SplitWindowDifference(self):
        """
        Split Window Difference RGB (greyscale):
        (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_SplitWindowDifference.pdf>`__ for reference)

        .. image:: /_static/SplitWindowDifference.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        data = ds["CMI_C15"].data - ds["CMI_C13"].data

        # _normalize values
        data = _normalize(data, -10, 10)

        # The final RGB array :)
        RGB = np.dstack([data, data, data])

        ds["SplitWindowDifference"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["SplitWindowDifference"].attrs[
            "Quick Guide"
        ] = "http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_SplitWindowDifference.pdf"
        ds["SplitWindowDifference"].attrs["long_name"] = "Split Window Difference"

        return ds["SplitWindowDifference"]

    def NightFogDifference(self):
        """
        Night Fog Difference RGB (greyscale):
        (See `Quick Guide <http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_NightFogBTD.pdf>`__ for reference)

        .. image:: /_static/NightFogDifference.png


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        data = ds["CMI_C13"].data - ds["CMI_C07"].data

        # _normalize values
        data = _normalize(data, -90, 15)

        # Invert data
        data = 1 - data

        # The final RGB array :)
        RGB = np.dstack([data, data, data])

        ds["NightFogDifference"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["NightFogDifference"].attrs[
            "Quick Guide"
        ] = "http://cimss.ssec.wisc.edu/goes/OCLOFactSheetPDFs/ABIQuickGuide_NightFogBTD.pdf"
        ds["NightFogDifference"].attrs["long_name"] = "NightFogDifference"

        return ds["NightFogDifference"]

    def RocketPlume(self, night=False):
        """
        Rocket Plume RGB

        For identifying rocket launches.

        See `this blog <https://cimss.ssec.wisc.edu/satellite-blog/archives/41335>`__ and
        the `Quick Guide <https://cimss.ssec.wisc.edu/satellite-blog/images/2021/06/QuickGuide_Template_GOESRBanner_Rocket_Plume.pdf>`__
        for reference

        .. image:: /_static/RocketPlume.png

        Parameters
        ----------
        night : bool
            If the area is in night, turn this on to use a different channel
            than the daytime application.


        """
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        R = ds["CMI_C07"].data
        G = ds["CMI_C08"].data
        if not night:
            B = ds["CMI_C02"].data
        else:
            B = ds["CMI_C05"].data

        # _normalize values
        R = _normalize(R, 273, 338)
        G = _normalize(G, 233, 253)
        B = _normalize(B, 0, 0.80)

        # The final RGB array :)
        RGB = np.dstack([R, G, B])

        ds["RocketPlume"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["RocketPlume"].attrs[
            "Quick Guide"
        ] = "https://cimss.ssec.wisc.edu/satellite-blog/images/2021/06/QuickGuide_Template_GOESRBanner_Rocket_Plume.pdf"
        ds["RocketPlume"].attrs["long_name"] = "Rocket Plume"

        return ds["RocketPlume"]

    def NormalizedBurnRatio(self):
        """
        Normalized Burn Ratio

        **THIS FUNCTION IS NOT FULLY DEVELOPED. Need more info.**

        NBR= (0.86 µm – 2.2 µm)/(0.86 um + 2.2 um)


        https://ntrs.nasa.gov/citations/20190030825

        """
        warnings.warn(
            "THE `NormalizedBurnRatio` FUNCTION IS NOT FULLY DEVELOPED. NEED MORE INFO."
        )
        ds = self._obj

        # Load the three channels into appropriate R, G, and B variables
        C3 = ds["CMI_C03"].data
        C6 = ds["CMI_C06"].data
        data = (C3 - C6) / (C3 + C6)

        # Invert data
        # data = 1-data

        # The final RGB array :)
        RGB = np.dstack([data, data, data])

        ds["NormalizedBurnRatio"] = (("y", "x", "rgb"), RGB)
        ds["rgb"] = ["R", "G", "B"]
        ds["NormalizedBurnRatio"].attrs[
            "Quick Guide"
        ] = "https://ntrs.nasa.gov/citations/20190030825"
        ds["NormalizedBurnRatio"].attrs["long_name"] = "Normalized Burn Ratio"

        return ds["NormalizedBurnRatio"]



from pathlib import Path
from setuptools import setup, find_packages

HERE = Path(__file__).parent
README = (HERE / 'README.md').read_text(encoding="utf8")

setup(
    name = 'goes2go',
    version = '0.0.4',
    author = 'Brian K. Blaylock',
    author_email = "blaylockbk@gmail.com",
    description = 'Retreive GOES-16/17 data from AWS. Also proves some RGB recipes.',
    long_description = README,
    long_description_content_type = 'text/markdown',
    project_urls = {
        'Source Code': 'https://github.com/blaylockbk/goes2go',
        'Documentation': 'https://blaylockbk.github.io/goes2go/_build/html/',
    },
    license = "MIT",
    packages = find_packages(),
    package_data = {
        "": ['*.cfg'],
    },
    #python_requires='>3.7',   # Will this requirement cause problems for people??
    install_requires = ['numpy', 'pandas', 'xarray', 's3fs', 'netcdf4'],
    keywords = ['xarray', 'meteorology', 'weather', 'GOES', 'forecast'],
    classifiers = [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
    zip_safe = False,
)

###############################################################################
## Brian's Note: How to upload a new version to PyPI
## -------------------------------------------------
# Created a new conda environment with twine
# conda create -n pypi python=3 twine pip -c conda-forge
'''
conda activate pypi
cd goes2go
python setup.py sdist bdist_wheel
twine check dist/*
# Test PyPI
twine upload --skip-existing --repository-url https://test.pypi.org/legacy/ dist/*
# PyPI
twine upload --skip-existing dist/*
'''
